' Scroll16b-01.bas (PIC16F628A, 4 MHz, PDS 3.5.4.5)
' Copyright (c) by zuisti, all rights reserved!

'Version 0.1, 967 (C: 1046) words, 55.3 (46.8) Hz refresh rate
' pure basic (best portable)
' array[index] access method is used (no pointer)

        Device = 16F628A 
        Xtal = 4
        Optimiser_Level = 3

    Declare FSR_CONTEXT_SAVE  Off

    On_Hardware_Interrupt GoTo Interrupt_Routine

$define SCROLLTIME 6 ' loop count value, the smaller is faster (fastest: 1)

' hw defines:
$define COMMON_CATHODE       'common cathode rows, common anode columns !!!

$define DATA_PIN  PORTB.3 ' the hc595 data

'Store pin: PORTA.3  (RA3), positive clock for shiftregs output storage
$define StorePulse PORTA.3 = 1 : PORTA.3 = 0

'clock pin: PORTB.0  (RB0), positive clock for shift in
$define ClockPulse PORTB.0 = 1 : PORTB.0 = 0

$define SCROLL_PIN PORTB.4   ' a jumper for (re)start (open) / stop (closed) scrolling

'Line decoder inputs: PORTA<2:0>, RA0 = LSB
$define DisableLines PORTA = 0b0111  ' = 7, all 1 (the outputs are active low)

'moved to the end
'Include "bitmap7.inc" ' ASCII font bitmap table, 7 bytes/char

'moved to the end
'message: LData "This is a very-very long message :-) ",0 ' end sign

Dim pattbuf[7]      As Byte ' holds the actual char's line patterns
Dim linebuf[7*7]    As Byte 
' 7 bytes for the first (uppermost) line (row), 7 bytes for the 2nd, ...

Dim ch              As Byte
Dim rowcnt          As Byte
Dim bytecnt         As Byte
Dim bitcnt          As Byte
Dim pat_idx         As Byte
Dim buf_idx         As Byte
Dim carry_over      As Byte
Dim width           As Byte
Dim scrollcnt       As Byte

Dim bitmap_idx      As Word
Dim msg_idx         As Word

Symbol TMR2IF = PIR1.1   'timer2 IT flag (bank 0)
Symbol TMR2ON = T2CON.2  'timer2 on/off (bank 0)


' main:

  GoSub HW_Init
  GoSub Clear_Linebuf
  msg_idx = 0 ' first msg char

  'For prompt scroll:
  width = 0
  scrollcnt = 1 

  While 1=1
    If SCROLL_PIN = 1 Then ' sum max 1685 us
       Dec scrollcnt
       If  scrollcnt = 0 Then
           scrollcnt = SCROLLTIME
           If width = 0 Then
              GoSub NextChar        ' put it to ch, 31 (C: 48) us
              width = 6
              If ch = 32 Then width = 3
              GoSub Get_Patterns    ' of ch, 296 (C: 435) us
           EndIf
           Dec width
           GoSub Shift_In_A_Column  ' 1339 (1656) us !!!!
       EndIf
    EndIf  
    GoSub Display ' 6.08 ms (without timer2 waiting !) 6.08/7 = appr 870 us / row
  Wend

' subroutines

HW_Init:
    CMCON = 0b00000111 ' switch off comparators
    DisableLines       ' PORTA = 7, also store pin is zeroed
    PORTB = 0          ' also data and clock pins are zeroed
    TRISA = 0          ' PORTA pins are outputs
    TRISB = 0xF2       ' only PORTB.0,2,3 are outputs
    Clear OPTION_REG.7 ' RBPU, portb pullups on 

' init timer2 For Line timing
'    one row display = 870 us (without timer2 waiting) but
'    with scroll (worst case = 19+31+296+1339 = 1.685 ms) 
'    sum = 1685 + 870 = 2555 us (min. for timer2)

'    Prescaler=1:1 PostScaler=1:12 Period = 2.556 ms
'    4 MHz: 2.56 (3.028) ms/line, 18.058 (21.35) ms the loop, 55.3 (46.8) Hz refresh rate

    PR2 = 213           ' Timer2 Match value
    T2CON = 0b01011100  ' enable timer2 with 1:12 postscaler, 1:1 prescaler
    TMR2IF = 0          ' clear IT flag
    PIE1 = 2            ' only TMR2IE = 1: enable Timer2 IT

    INTCON = 0xC0 ' Enable global (GIE) and periferial (PEIE) ITs
 Return ' HW_Init

Interrupt_Routine:
    Context Save
    DisableLines   ' PORTA = 7
    TMR2ON = 0 ' disable timer2
    TMR2IF = 0 ' clear IT flag
  Context Restore

Get_Patterns: ' input: character in ch
    ch = ch - 32 ' the space is the first char in the bitmap table
    'calculate the bitmap table offset of the first pattern of the char ch
    bitmap_idx = 8*ch - ch ' = 7*ch, but faster
    
    rowcnt = 0 'we have 7 patterns (0...6)
    Repeat
        ' read pattern from the bitmap table
        ch = LRead8 bitmap[bitmap_idx + rowcnt] 
        ' drop unused two bits (7,6)
        ch = ch << 2 
        $ifdef COMMON_CATHODE
            ch = ~ch ' inverted (0<>1)
        $endif
        pattbuf[rowcnt] = ch ' and store it
        Inc rowcnt
    Until rowcnt = 7
  Return ' Get_Patterns

NextChar:
    While 1=1
        ch = LRead8 message[msg_idx]
        Inc msg_idx
        If ch <> 0 Then Break   ' not 0 (not end sign)
        msg_idx = 0             ' else back to begin
    Wend
  Return ' character in ch
     
Shift_In_A_Column:
    pat_idx = 6  ' last (7th, lowermost) pattern
    buf_idx = 48 ' last (7th, lowermost) line's last (7th) byte
    rowcnt = 7
    Repeat 
        ' unfortunately (in PDS) no rotate with carry
        
        ' pattern rotate left with carry
        ch = pattbuf[pat_idx]
        carry_over = ch & 0x80
        ch = ch << 1 ' a 0 shifted in from right but not used
        pattbuf[pat_idx] = ch
        Dec pat_idx
        
        ' rotate a whole Line (7 bytes) left with carry
        bytecnt = 7
        Repeat
            ch = linebuf[buf_idx]
            ch = ch << 1
            If carry_over <> 0 Then ch.0 = 1
            carry_over = STATUS & 1 ' equal to the value of the carry flag. This is the repair !!
            linebuf[buf_idx] = ch   ' carry is destroyed here !
            Dec buf_idx
'            carry_over = ch & 0x80 '  !! fixed and moved up
            Dec bytecnt 
        Until bytecnt = 0
        Dec rowcnt
    Until rowcnt = 0
  Return ' Shift_In_A_Column

Display: 
    buf_idx = 0 'from begin of buffer
    ' loop until rowcnt = 7
    rowcnt = 0
    Repeat ' 870 us / row
        bytecnt = 7
        Repeat
            ch = linebuf[buf_idx]
            Inc buf_idx
            bitcnt = 8
            Repeat
                DATA_PIN = ch.7
                ch = ch << 1
                ClockPulse
                Dec bitcnt
            Until bitcnt = 0
            Dec bytecnt
        Until bytecnt = 0
        
        'timed display (same row brightness)
        Repeat Until TMR2ON = 0 'wait for timer2 IT (timer2 and all lines off)
        TMR2ON = 1 'restart timer2, also a time gap to switch off drivers
        StorePulse
        PORTA = rowcnt ' switch on the actual line driver
        Inc rowcnt
    Until rowcnt = 7
  Return ' Display

Clear_Linebuf:
    rowcnt = 49
    Repeat
        $ifdef COMMON_CATHODE
            linebuf[rowcnt-1] = 0
        $else
            linebuf[rowcnt-1] = 255
        $endif
        Dec rowcnt
    Until rowcnt = 0
  Return

' -------

message: LData "This is a very-very long message :-) ",0 ' end sign

Include "bitmap7.inc" ' ASCII font bitmap table, 7 bytes/char

'------ End
