' Scroll16b-03.bas (PIC16F628A, 4 MHz, PDS 3.5.4.5)
' Copyright (c) by zuisti, all rights reserved!

'Version 0.1, 967 (C: 1046) words, 55.3 (46.8) Hz refresh rate
' pure basic (best portable)
' array[index] access method is used (no pointer)

'version 0.2 not converted to basic because no pointers in PDS

'version 0.3, 955 (C: 1010) words, 70 (60.4) Hz refresh rate (a significant improvement)
' the same as 0.2 in C but indirect array access by FSR and INDF (legal in PDS)
' FSR is a char pointer (hardware char *), and INDF is the *FSR (hw too)
' only one FSR exists in a PIC16 (except the enhanced types) but three in PIC18


        Device = 16F628A 
        Xtal = 4
        Optimiser_Level = 3

    Declare FSR_CONTEXT_SAVE  Off

    On_Hardware_Interrupt GoTo Interrupt_Routine

$define SCROLLTIME 6 ' loop count value, the smaller is faster (fastest: 1)

' hw defines:
$define COMMON_CATHODE       'common cathode rows, common anode columns !!!

$define DATA_PIN  PORTB.3 ' the hc595 data

'Store pin: PORTA.3  (RA3), positive clock for shiftregs output storage
$define StorePulse PORTA.3 = 1 : PORTA.3 = 0

'clock pin: PORTB.0  (RB0), positive clock for shift in
$define ClockPulse PORTB.0 = 1 : PORTB.0 = 0

$define SCROLL_PIN PORTB.4   ' a jumper for (re)start (open) / stop (closed) scrolling

'Line decoder inputs: PORTA<2:0>, RA0 = LSB
$define DisableLines PORTA = 0b0111  ' = 7, all 1 (the outputs are active low)

'moved to the end
'Include "bitmap7.inc" ' ASCII font bitmap table, 7 bytes/char

'moved to the end
'message: LData "This is a very-very long message :-) ",0 ' end sign

Dim pattbuf[7]      As Byte ' holds the actual char's line patterns
Dim linebuf[7*7]    As Byte 
' 7 bytes for the first (uppermost) line (row), 7 bytes for the 2nd, ...

Dim ch              As Byte
Dim rowcnt          As Byte
Dim bytecnt         As Byte
Dim bitcnt          As Byte
Dim pat_fsr         As Byte
Dim buf_fsr         As Byte
Dim carry_over      As Byte
Dim width           As Byte
Dim scrollcnt       As Byte

Dim bitmap_idx      As Word
Dim msg_idx         As Word

Symbol TMR2IF = PIR1.1   'timer2 IT flag (bank 0)
Symbol TMR2ON = T2CON.2  'timer2 on/off (bank 0)


' main:

  GoSub HW_Init
  GoSub Clear_Linebuf
  msg_idx = 0 ' first msg char

  'For prompt scroll:
  width = 0
  scrollcnt = 1 

  While 1=1
    If SCROLL_PIN = 1 Then ' sum max 1152 us
       Dec scrollcnt
       If  scrollcnt = 0 Then
           scrollcnt = SCROLLTIME
           If width = 0 Then
              GoSub NextChar        ' put it to ch, 31 us
              width = 6
              If ch = 32 Then width = 3
              GoSub Get_Patterns    ' of ch, 277 us
           EndIf
           Dec width
           GoSub Shift_In_A_Column  ' 825 us !!!!
       EndIf
    EndIf  
    GoSub Display ' 6.02 ms (without timer2 waiting !) 6.02/7 = appr 860 us / row
  Wend

' subroutines

HW_Init:
    CMCON = 0b00000111 ' switch off comparators
    DisableLines       ' PORTA = 7, also store pin is zeroed
    PORTB = 0          ' also data and clock pins are zeroed
    TRISA = 0          ' PORTA pins are outputs
    TRISB = 0xF2       ' only PORTB.0,2,3 are outputs
    Clear OPTION_REG.7 ' RBPU, portb pullups on 

' init timer2 For Line timing
'    one row display = 860 us (without timer2 waiting) but
'    with scroll (worst case = 19+31+277+825 = 1152 us) 
'    sum = 1152 + 860 = 2012 us (min. for timer2)

'    Prescaler=1:1 PostScaler=1:12 Period = 2.016 ms
'    4 MHz: 2.016 ms/line, 14.28 ms the loop, 70 Hz refresh rate

    PR2 = 168           ' Timer2 Match value
    T2CON = 0b01011100  ' enable timer2 with 1:12 postscaler, 1:1 prescaler
    TMR2IF = 0          ' clear IT flag
    PIE1 = 2            ' only TMR2IE = 1: enable Timer2 IT

    INTCON = 0xC0 ' Enable global (GIE) and periferial (PEIE) ITs
 Return ' HW_Init

Interrupt_Routine:
    Context Save
    DisableLines   ' PORTA = 7
    TMR2ON = 0 ' disable timer2
    TMR2IF = 0 ' clear IT flag
  Context Restore

Get_Patterns: ' input: character in ch
    FSR = VarPtr pattbuf ' points the begin (patt#0)
    ch = ch - 32 ' the space is the first char in the bitmap table
    'calculate the bitmap table offset of the first pattern of the char ch
    bitmap_idx = 8*ch - ch ' = 7*ch, but faster
    
    rowcnt = 0 'we have 7 patterns (0...6)
    Repeat
        ' read pattern from the bitmap table
        ch = LRead8 bitmap[bitmap_idx + rowcnt] 
        ' drop unused two bits (7,6)
        ch = ch << 2 
        ' put it to array
        $ifdef COMMON_CATHODE
            INDF = ~ch ' inverted (0<>1)
        $else
            INDF = ch
        $endif
        Inc FSR
        Inc rowcnt
    Until rowcnt = 7
  Return ' Get_Patterns

NextChar:
    While 1=1
        ch = LRead8 message[msg_idx]
        Inc msg_idx
        If ch <> 0 Then Break   ' not 0 (not end sign)
        msg_idx = 0             ' else back to begin
    Wend
  Return ' character in ch
     
Shift_In_A_Column:
    pat_fsr = VarPtr pattbuf
    pat_fsr = pat_fsr + 6  ' last (7th) pattern
    buf_fsr = VarPtr linebuf
    buf_fsr = buf_fsr + 48; ' global, points the last line's last byte
    ' because there is only one FSR
    rowcnt = 7
    Repeat 
        ' unfortunately (in PDS) no rotate with carry
        
        ' pattern rotate left with carry
        FSR = pat_fsr
        Dec pat_fsr
        carry_over = INDF & 0x80
        INDF = INDF << 1 ' a 0 shifted in from right but not used
        
        ' rotate a whole line (7 bytes) left with carry, from buf_fsr, FSR is destroyed
        FSR = buf_fsr ' first points the buffer end
        bytecnt = 7
        Repeat
            INDF = INDF << 1
            If carry_over <> 0 Then INDF.0 = 1
'            carry_over = INDF & 0x80
            carry_over = STATUS & 1 ' equal to the value of the carry flag. This is the repair !!
            Dec FSR
            Dec bytecnt 
        Until bytecnt = 0
        buf_fsr = FSR
        Dec rowcnt
    Until rowcnt = 0
  Return ' Shift_In_A_Column

Display: 
    FSR = VarPtr linebuf ' points to the first byte    
    ' loop until rowcnt = 7
    rowcnt = 0
    Repeat ' 870 us / row
        bytecnt = 7
        Repeat
            ch = INDF
            bitcnt = 8
            Repeat
                DATA_PIN = ch.7
                ch = ch << 1
                ClockPulse
                Dec bitcnt
            Until bitcnt = 0
            Inc FSR
            Dec bytecnt
        Until bytecnt = 0
        
        'timed display (same row brightness)
        Repeat Until TMR2ON = 0 'wait for timer2 IT (timer2 and all lines off)
        TMR2ON = 1 'restart timer2, also a time gap to switch off drivers
        StorePulse
        PORTA = rowcnt ' switch on the actual line driver
        Inc rowcnt
    Until rowcnt = 7
  Return ' Display

Clear_Linebuf:
    FSR = VarPtr linebuf
    FSR = FSR + 48 ' last byte
    rowcnt = 49
    Repeat
        $ifdef COMMON_CATHODE
            INDF = 0
        $else
            INDF = 255
        $endif
        Dec FSR
        Dec rowcnt
    Until rowcnt = 0
  Return

' -------

message: LData "This is a very-very long message :-) ",0 ' end sign

Include "bitmap7.inc" ' ASCII font bitmap table, 7 bytes/char

'------ End
